/**
 * NOTE:
 *
 * HAL_TCP_xxx API reference implementation: wrappers/os/ubuntu/HAL_TCP_linux.c
 *
 */
#include "infra_types.h"
#include "infra_defs.h"
#include "infra_compat.h"
#include "wrappers_defs.h"

#include <winsock2.h>
#include <Windows.h>
#include <ws2tcpip.h>

#include <openssl/ssl.h>
#include <openssl/err.h>

/**
 * @brief Deallocate memory block
 *
 * @param[in] ptr @n Pointer to a memory block previously allocated with platform_malloc.
 * @return None.
 * @see None.
 * @note None.
 */
void HAL_Free(void *ptr)
{
	if(ptr)
		free(ptr);
}


/**
 * @brief Get device name from user's system persistent storage
 *
 * @param [ou] device_name: array to store device name, max length is IOTX_DEVICE_NAME_LEN
 * @return the actual length of device name
 */
int HAL_GetDeviceName(char device_name[IOTX_DEVICE_NAME_LEN + 1])
{
	memset(device_name,0,IOTX_DEVICE_NAME_LEN + 1);
	return(snprintf(device_name,IOTX_DEVICE_NAME_LEN,"%s","nHGjLvGlC7IKcGJRYCa1"));
}


/**
 * @brief Get device secret from user's system persistent storage
 *
 * @param [ou] device_secret: array to store device secret, max length is IOTX_DEVICE_SECRET_LEN
 * @return the actual length of device secret
 */
int HAL_GetDeviceSecret(char device_secret[IOTX_DEVICE_SECRET_LEN + 1])
{
	memset(device_secret,0,IOTX_DEVICE_SECRET_LEN + 1);
	return(snprintf(device_secret,IOTX_DEVICE_SECRET_LEN,"%s","9KUecjotVKoOtekJacpAoYRHW2GL04Bi"));
}


/**
 * @brief Get firmware version
 *
 * @param [ou] version: array to store firmware version, max length is IOTX_FIRMWARE_VER_LEN
 * @return the actual length of firmware version
 */
int HAL_GetFirmwareVersion(char *version)
{
	return (int)0;
}


/**
 * @brief Get product key from user's system persistent storage
 *
 * @param [ou] product_key: array to store product key, max length is IOTX_PRODUCT_KEY_LEN
 * @return  the actual length of product key
 */
int HAL_GetProductKey(char product_key[IOTX_PRODUCT_KEY_LEN + 1])
{
	memset(product_key,0,IOTX_PRODUCT_KEY_LEN + 1);
	return(snprintf(product_key,IOTX_PRODUCT_KEY_LEN,"%s","a1a840P0BS3"));
}


/**
 * @brief Allocates a block of size bytes of memory, returning a pointer to the beginning of the block.
 *
 * @param [in] size @n specify block size in bytes.
 * @return A pointer to the beginning of the block.
 * @see None.
 * @note Block value is indeterminate.
 */
void *HAL_Malloc(uint32_t size)
{
	if(size==0)
		return(NULL);
	return malloc(size);
}


/**
 * @brief Create a mutex.
 *
 * @retval NULL : Initialize mutex failed.
 * @retval NOT_NULL : The mutex handle.
 * @see None.
 * @note None.
 */
void *HAL_MutexCreate(void)
{
	HANDLE mutex;

	if (NULL == (mutex = CreateMutex(NULL, FALSE, NULL))) {
	    printf("create mutex error");
	}

	return mutex;
}


/**
 * @brief Destroy the specified mutex object, it will release related resource.
 *
 * @param [in] mutex @n The specified mutex.
 * @return None.
 * @see None.
 * @note None.
 */
void HAL_MutexDestroy(void *mutex)
{
	if (0 == CloseHandle(mutex)) {
		printf("destroy mutex error");
	};
}


/**
 * @brief Waits until the specified mutex is in the signaled state.
 *
 * @param [in] mutex @n the specified mutex.
 * @return None.
 * @see None.
 * @note None.
 */
void HAL_MutexLock(void *mutex)
{
	if (WAIT_FAILED == WaitForSingleObject(mutex, INFINITE)) {
		printf("lock mutex error");
	}
}


/**
 * @brief Releases ownership of the specified mutex object..
 *
 * @param [in] mutex @n the specified mutex.
 * @return None.
 * @see None.
 * @note None.
 */
void HAL_MutexUnlock(void *mutex)
{
	ReleaseMutex(mutex);
}


/**
 * @brief Writes formatted data to stream.
 *
 * @param [in] fmt: @n String that contains the text to be written, it can optionally contain embedded format specifiers
     that specifies how subsequent arguments are converted for output.
 * @param [in] ...: @n the variable argument list, for formatted and inserted in the resulting string replacing their respective specifiers.
 * @return None.
 * @see None.
 * @note None.
 */
void HAL_Printf(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vprintf(fmt, args);
	va_end(args);

	fflush(stdout);
}


/**
 * @brief Sleep thread itself.
 *
 * @param [in] ms @n the time interval for which execution is to be suspended, in milliseconds.
 * @return None.
 * @see None.
 * @note None.
 */
void HAL_SleepMs(uint32_t ms)
{
	Sleep(ms);
}


/**
 * @brief Writes formatted data to string.
 *
 * @param [out] str: @n String that holds written text.
 * @param [in] len: @n Maximum length of character will be written
 * @param [in] fmt: @n Format that contains the text to be written, it can optionally contain embedded format specifiers
     that specifies how subsequent arguments are converted for output.
 * @param [in] ...: @n the variable argument list, for formatted and inserted in the resulting string replacing their respective specifiers.
 * @return bytes of character successfully written into string.
 * @see None.
 * @note None.
 */
int HAL_Snprintf(char *str, const int len, const char *fmt, ...)
{
	int ret;
	va_list args;

	va_start(args, fmt);
	ret = _vsnprintf(str, len - 1, fmt, args);
	va_end(args);

	return ret;
}

#define MAX_SSL_CONNECTIONS	5
typedef struct{
	SSL_METHOD *pMethod;
	SSL_CTX *pCtx;
	SSL* pSsl;
	int fd;
}sslConnDef_t;

static int openSslInited = 0 ;
static sslConnDef_t* connArray[MAX_SSL_CONNECTIONS] = {0};

static void init_open_ssl()
{
	WSADATA wsaData;

	if(openSslInited)
		return;

	WSAStartup(0x202, &wsaData);
	SSL_library_init();

	openSslInited = 1;
}

int32_t HAL_SSL_Destroy(uintptr_t handle)
{
	int index = (int)handle -1;

	printf("Enter HAL_SSL_Destroy\n\r");
	if((handle== 0)||(handle>MAX_SSL_CONNECTIONS))
		return(0);
	if(connArray[index]->pSsl){
		SSL_shutdown(connArray[index]->pSsl);
		SSL_free(connArray[index]->pSsl);
	}
	if(connArray[index]->pCtx)
		SSL_CTX_free(connArray[index]->pCtx);
	if(connArray[index]->fd > 0)
		closesocket(connArray[index]->fd);
	free(connArray[index]);
	connArray[index] = NULL;

	return(0);
}

static int tcp_open(const char *host, uint16_t port)
{
	int sockfd,res,err, flag=0;
	unsigned long nonBlock=1;
	struct hostent *hp;
	struct sockaddr_in addrServer;


	sockfd = socket(AF_INET, SOCK_STREAM, 0);   /* socket */
	hp = gethostbyname(host);
	if(hp==NULL)
		return(-1);

	memset(&addrServer, 0, sizeof(addrServer));
	memcpy(&(addrServer.sin_addr), hp->h_addr, hp->h_length);
	addrServer.sin_family = AF_INET;
	addrServer.sin_port = htons((unsigned short)port);

	printf("connecting to %s", host);
	ioctlsocket(sockfd,FIONBIO,&nonBlock);
	while(1){
		flag = 0;
		res = connect(sockfd, (struct sockaddr *)&addrServer, sizeof(struct sockaddr));
		if(res==0)
			break;
		else{
			err = WSAGetLastError();
			switch(err){
				case(WSAEWOULDBLOCK):
					printf("tcp_open: WSAEWOULDBLOCK");
					Sleep(100);
				break;

				case(WSAEISCONN):
					flag=1;
				break;

				default:
					printf("tcp_open error %d", err);
					closesocket(sockfd);
					return(-1);
				break;
			}
			if(flag==1)
				break;
			else{
				Sleep(100);
			}
		}
	}

	printf("\n TCP connect successfully!\n");
	return sockfd;
}

uintptr_t HAL_SSL_Establish(const char *host, uint16_t port, const char *ca_crt, uint32_t ca_crt_len)
{
	SSL_METHOD *method;
	SSL_CTX *ctx;
	int i,fd=0,err,res;

	printf("Enter HAL_SSL_Establish\n\r");
	init_open_ssl();

	for(i=0;i<MAX_SSL_CONNECTIONS;i++)
		if(connArray[i]==NULL)
			break;
	if(i>=MAX_SSL_CONNECTIONS){
		printf("Exceed connection limit\n\r");
		return(0);
	}
	if((fd=tcp_open(host,port))<0){
		printf("tcp_open failed\n\r");
		return(0);
	}

	method = SSLv23_client_method();
	ctx = SSL_CTX_new(method);
	if (ctx == NULL){
		closesocket(fd);
		return(0);
	}

	connArray[i] = (sslConnDef_t *)malloc(sizeof(sslConnDef_t));
	if(connArray[i] == NULL)
		goto ErrRet;
	memset(connArray[i],0,sizeof(sslConnDef_t));
	connArray[i]->pMethod = method;
	connArray[i]->pCtx = ctx;
	connArray[i]->fd = fd;

	SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
	connArray[i]->pSsl = SSL_new (ctx);
	SSL_set_fd (connArray[i]->pSsl, fd);
	SSL_set_connect_state(connArray[i]->pSsl);

	while(1){
		res = SSL_connect (connArray[i]->pSsl);
		if(res>0)
			break;
		printf("HAL_SSL_Establish: res %d\n", res);
		err = SSL_get_error(connArray[i]->pSsl, res);
		if ((err == SSL_ERROR_WANT_WRITE) || (err == SSL_ERROR_WANT_READ)){
			Sleep(50);
			continue;
		}
		else{
			goto ErrRet;
		}
	}

	printf("HAL_SSL_Establish successfully\n\r");
	return(i+1);

ErrRet:
	if(connArray[i]->pSsl)
		SSL_free(connArray[i]->pSsl);
	if(connArray[i]->pCtx)
		SSL_CTX_free(connArray[i]->pCtx);
	if(fd>0)
		closesocket(fd);
	connArray[i] = NULL;
	printf("HAL_SSL_Establish failed\n\r");
	return(0);
}


int HAL_SSL_Read(uintptr_t handle, char *buf, int len, int timeout_ms)
{
	int res, errCode,index;

	if((handle== 0)||(handle>MAX_SSL_CONNECTIONS)){
		printf("Invalid handle %d\n\r",(int)handle);
		return(0);
	}
	index = (int)handle - 1;

	while(1){
		res = SSL_read(connArray[index]->pSsl,buf,len);
		if(res>0)
			break;
		else{
			errCode = SSL_get_error(connArray[index]->pSsl, res);
			switch(errCode){
				case(SSL_ERROR_WANT_CONNECT):
				case(SSL_ERROR_SYSCALL):
				case(SSL_ERROR_SSL):
					printf("HAL_SSL_Read: connection error\n\r");
					return(-2);
				break;

				case(SSL_ERROR_WANT_READ ):
					Sleep(100);
					return(0);
				break;

				default:
					printf("errCode %d\n\r",errCode);
					return(-1);
				break;
			}
		}
	}

	printf("Rcved %d B from server: %s\n\r",res, buf);
	return(res);
}


int HAL_SSL_Write(uintptr_t handle, const char *buf, int len, int timeout_ms)
{
	int res, errCode, index = (int)handle-1;
	int contFlag, sent=0;
	unsigned char *pStart = (unsigned char *)buf;

	if((handle== 0)||(handle>MAX_SSL_CONNECTIONS))
		return(0);

	printf("HAL_SSL_Write %d,len %d,timeout %d\n\r",(int)handle,len,timeout_ms);

	while(1){
		contFlag = 0;
		res = SSL_write(connArray[index]->pSsl,pStart,len-sent);
		if(res > 0){
			sent += res;
			pStart += res;
			printf("sent %d B\n\r", sent);
			if(sent<len)
				contFlag=1;
		}else{
			errCode = SSL_get_error(connArray[index]->pSsl, res);
			printf("res %d, errcode %d\n\r",res, errCode);
			switch(errCode){
				case(SSL_ERROR_WANT_READ):
				case(SSL_ERROR_WANT_WRITE):
					contFlag =1;
					Sleep(50);
				break;

				case(SSL_ERROR_ZERO_RETURN):
				case(SSL_ERROR_SYSCALL):
				case(SSL_ERROR_SSL):
					printf("HAL_SSL_Write: connection error\n\r");
					return(-1);
				break;

				default:
					printf("HAL_SSL_Write error, errcode %d\n\r",errCode);
					contFlag = 1;
					Sleep(50);
				break;
			}
		}
		if(contFlag==0)
			break;
		else
			Sleep(100);
	}

	printf("Send msg (%d/%d)%s\n\r",len,sent,buf);
	return(sent);
}


/**
 * @brief Retrieves the number of milliseconds that have elapsed since the system was boot.
 *
 * @return the number of milliseconds.
 * @see None.
 * @note None.
 */
uint64_t HAL_UptimeMs(void)
{
	return (uint64_t)(GetTickCount());
}


